import logging
import multiprocessing
import asyncio
import random

import psutil

from utility import is_linux


class SystemMonitor(object):
    """ Utility functions to read Raspberry pi system state and Sense Hat data asynchronously """
    def __init__(self):
        self.logger = logging.getLogger("SystemMonitor")
        random.seed()
        self.sensors = {
                            'temp': 50.0,
                            'pressure': 1.0,
                            'humidity': .1,
                            'pitch': 0,
                            'roll': 0,
                            'yaw': 0
                        }

    async def get_system_data(self):
        return {
                    'free_space': psutil.disk_usage("/").free / 1024 / 1024,
                    'cpu_temp': "N/A" if not is_linux() else psutil.sensors_temperatures()[0],
                    'cpu_load': psutil.cpu_percent()
                }

    async def get_sensors(self):
        self.sensors['temp'] += random.uniform(-.1, .1)
        self.sensors['pressure'] += random.uniform(-.1, .1)
        self.sensors['humidity'] += random.uniform(-.1, .1)
        self.sensors['pitch'] += random.uniform(-.1, .1)
        self.sensors['roll'] += random.uniform(-.1, .1)
        self.sensors['yaw'] += random.uniform(-.1, .1)
        await asyncio.sleep(0.2)
        return self.sensors




import shutil
from pathlib import Path
import os
    
def setup_startup(name="edurov"):
    system_directory = Path(__file__).parent
    systemd_user_dir = Path("/lib/systemd/system/")

    os.makedirs(systemd_user_dir, exist_ok=True)
    user = os.getlogin()

    with open(system_directory / "pyedurov2.service") as source:
        with open(systemd_user_dir / "pyedurov2.service", "w") as target:
            target.write(source.read().format(user_name = user, advertising_name=name))

    os.system("systemctl daemon-reload")
    os.system("systemctl enable pyedurov2.service")

    print("pyedurov2 service added to startup.")



import base64
import logging
import multiprocessing
import asyncio
import time

import websockets

from ..hardware import camera
from ..utility import get_host_ip


class CameraServer(multiprocessing.Process):
    """ Creates a new process that Exposes the raspberry pi camera as a websocket image stream """
    def __init__(self, video_resolution='1024x768', fps=30, loglevel="INFO", port=8081):
        self.port = port
        self.video_resolution = [int(value) for value in video_resolution.split('x')]
        self.fps = fps
        self.loglevel = loglevel
        self.start_time = time.time()
        self.server = None
        self.ready = multiprocessing.Event()
        self.stop_event = multiprocessing.Event()

        super().__init__(target=self._runner, daemon=True)
        self.start()

    async def stop(self):
        self.stop_event.set()
        self.join()
        logging.debug("Camera process terminated")

    async def _wait_for_stop_event(self):
        while not self.stop_event.is_set():
            await asyncio.sleep(2)
        self.server.ws_server.close()

    async def _send_frames(self, websocket):
        if self.camera is None:
            return
        while not self.stop_event.is_set():
            with self.camera.stream.condition:
                self.camera.stream.condition.wait()
                if self.camera.stream.frame is not None:
                    await websocket.send("data:image/jpg;base64," + base64.b64encode(self.camera.stream.frame).decode())

    async def _handler(self, websocket, path):
        self.logger.info(f"Camera server received connection from {path}")
        start_cmd = await websocket.recv()
        send = asyncio.create_task(self._send_frames(websocket))
        await websocket.wait_closed()
        send.cancel()

    def _runner(self):
        asyncio.run(self._task())

    async def _task(self):
        logging.basicConfig(level=self.loglevel)
        self.logger = logging.getLogger("CameraServer")
        logging.getLogger("websockets.protocol").setLevel(logging.INFO)
        with camera.Camera(self.video_resolution, self.fps) as self.camera:
            self.server = websockets.serve(self._handler, get_host_ip(), self.port)
            stop_task = asyncio.create_task(self._wait_for_stop_event())

            self.logger.info(f"Camera websocket server started at ws://{get_host_ip()}:{self.port}")
            self.ready.set()

            await self.server
            await stop_task

        self.logger.info('Shutting down server')
        finish = time.time()
        seconds = finish - self.start_time
        self.logger.debug(f'Server was live for {seconds:.1f} seconds')
        
        if self.camera is not None:
            framerate = self.camera.stream.count / (finish - self.start_time)
        else:
            framerate = 0
        self.logger.debug(f'Sent {self.camera.stream.count} images at {framerate:.2f} fps')




import re
import socket
import subprocess
import sys
import threading
import io


class StreamingOutput(object):
    """A multithreaded output stream buffer, used for the camera server backend"""

    def __init__(self):
        self.frame = None
        self.buffer = io.BytesIO()
        self.condition = threading.Condition()
        self.count = 0

    def write(self, buf):
        if buf.startswith(b'\xff\xd8'):
            # New frame, copy the existing buffer's content and notify all
            # clients it's available
            self.buffer.truncate()
            with self.condition:
                self.frame = self.buffer.getvalue()
                self.condition.notify_all()
            self.buffer.seek(0)
            self.count += 1
        return self.buffer.write(buf)


def is_linux():
    return sys.platform.startswith("linux")


def is_windows():
    return sys.platform.startswith("win")


def is_osx():
    return sys.platform.startswith("darwin")


def is_raspberrypi():
    if not is_linux():
        return False
    try:
        with open('/sys/firmware/devicetree/base/model', 'r') as m:
            return 'raspberry pi' in m.read().lower()
    except Exception:
        pass
    return False


def get_host_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP



import json
import logging
import serial_asyncio
from serial.tools import list_ports
from ..utility import is_raspberrypi


class Arduino(object):
    """ Utility functions to perform Arduino communication asynchronously """
    def __init__(self, serial_port=None, baud_rate=115200):
        self.logger = logging.getLogger("Arduino")
        self.baud_rate = baud_rate

        if serial_port is None:
            self.logger.debug("Attempt to auto-detect Arduino serial port")
            coms = list_ports.comports()
            self.logger.debug(f"Available serial ports: {str([com.device for com in coms])}")
            arduino = [com for com in coms if com.manufacturer is not None and "Arduino" in com.manufacturer]
            if arduino:
                # We found an Arduino connected via USB
                serial_port = arduino[0].device
            elif is_raspberrypi():
                # We didn't find an Arduino on USB, but list_ports.comports 
                # does not list the hardware comport on most Raspberry pis, so find it manually.
                serial_port = "/dev/serial0"
            else:
                # We're not on a Raspberry Pi, and we didn't find any Raspberry PI, 
                # let's just try the first one available.
                serial_port = coms[0].device

        self.serial_port = serial_port
        self._reader = None
        self._writer = None

    async def open(self):
        self._reader, self._writer \
            = await serial_asyncio.open_serial_connection(url=self.serial_port, baudrate=self.baud_rate)

    async def close(self):
        self._writer.close()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def get_sensors(self):
        received = (await self._reader.readuntil(b"}"))
        self.logger.debug(f"Received data from Arduino: {received.decode('ascii', errors='ignore')}")
        try:
            return json.loads(received)
        except Exception as e:
            logging.error(e)
            return {}

    def set_actuators(self, values):
        message =   f"{int(values['starboard'])}," + \
                    f"{int(values['port'])}," + \
                    f"{int(values['vertical'])}," + \
                    f"{int(values['lights'])}," + \
                    f"{int(values['interval'])}," + \
                    f"{1 if values['armed'] else 0}\n"
        self.logger.debug(f"Sent data to Arduino: {message}")
        self._writer.write(message.encode("ascii"))



from setuptools import setup, find_packages
import pyedurov2

setup(

    name='pyedurov2',

    version=pyedurov2.__version__,

    description='Server package for Edu-ROV hardware v0.7',
    long_description='',

    url = 'https://github.com/simtind/edu-rover2-server',
    project_urls={
        'Source': 'https://github.com/simtind/edu-rover2-server',
        'Tracker': 'https://github.com/simtind/edu-rover2-server/issues',
    },

    author='Simen August Tinderholt',
    author_email='simtind@gmail.com',

    python_requires='>=3.7',

    classifiers=[

        'Development Status :: 5 - Production/Stable',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3.7',
    ],

    keywords='ROV education',

    install_requires=['picamera', 'sense-hat', 'pyserial-asyncio', 'asyncio', 'psutil', 'websockets', 'aiortc', 'aiohttp'],

    packages=find_packages(),
    package_data={
        'pyedurov2.system': ['*.service']
    },
    entry_points={
        'console_scripts': [
            'pyedurov2 = pyedurov2.server:run',
        ],
    },
)
